from .drive import (Drive, UploadError, DownloadError, RemoteNodeError,
                    FileConflictedError, InvalidNameError, ParentNotFoundError,
                    NodeNotFoundError, download_to_local,
                    download_to_local_by_id, upload_from_local,
                    upload_from_local_by_id)
from .network import AuthenticationError, NetworkError, ResponseError
from .cache import dict_from_node



import argparse
import asyncio
import concurrent.futures as cf
import contextlib as cl
import functools as ft
import hashlib
import io
import os
import os.path as op
import pathlib as pl
import sys

import yaml
import wcpan.logger as wl
import wcpan.worker as ww

from .drive import Drive, DownloadError, download_to_local, upload_from_local
from .util import stream_md5sum, get_default_conf_path


class AbstractQueue(object):

    def __init__(self, drive, jobs):
        self._drive = drive
        self._queue = ww.AsyncQueue(jobs)
        self._pool = None
        self._counter = 0
        self._table = {}
        self._total = 0
        self._failed = []
        self._raii = None

    async def __aenter__(self):
        async with cl.AsyncExitStack() as stack:
            self._pool = stack.enter_context(cf.ProcessPoolExecutor())
            self._raii = stack.pop_all()
        return self

    async def __aexit__(self, type_, exc, tb):
        await self._queue.shutdown()
        await self._raii.aclose()
        self._pool = None
        self._raii = None

    @property
    def drive(self):
        return self._drive

    @property
    def failed(self):
        return self._failed

    async def run(self, src_list, dst):
        if not src_list:
            return
        self._counter = 0
        self._table = {}
        total = (self.count_tasks(_) for _ in src_list)
        total = await asyncio.gather(*total)
        self._total = sum(total)
        for src in src_list:
            fn = ft.partial(self._run_one_task, src, dst)
            self._queue.post(fn)
        await self._queue.join()

    async def count_tasks(self, src):
        raise NotImplementedError()

    def source_is_folder(self, src):
        raise NotImplementedError()

    async def do_folder(self, src, dst):
        raise NotImplementedError()

    async def get_children(self, src):
        raise NotImplementedError()

    async def do_file(self, src, dst):
        raise NotImplementedError()

    def get_source_hash(self, src):
        raise NotImplementedError()

    async def get_source_display(self, src):
        raise NotImplementedError()

    async def _run_one_task(self, src, dst):
        self._update_counter_table(src)
        async with self._log_guard(src):
            if self.source_is_folder(src):
                rv = await self._run_for_folder(src, dst)
            else:
                rv = await self._run_for_file(src, dst)
        return rv

    async def _run_for_folder(self, src, dst):
        try:
            rv = await self.do_folder(src, dst)
        except Exception as e:
            wl.EXCEPTION('wcpan.drive.google', e)
            display = await self.get_source_display(src)
            self._add_failed(display)
            rv = None

        if not rv:
            return None

        children = await self.get_children(src)
        for child in children:
            fn = ft.partial(self._run_one_task, child, rv)
            self._queue.post(fn)

        return rv

    async def _run_for_file(self, src, dst):
        try:
            rv = await self.do_file(src, dst)
        except Exception as e:
            wl.EXCEPTION('wcpan.drive.google', e)
            display = await self.get_source_display(src)
            self._add_failed(display)
            rv = None
        return rv

    def _add_failed(self, src):
        self._failed.append(src)

    @cl.asynccontextmanager
    async def _log_guard(self, src):
        await self._log('begin', src)
        try:
            yield
        finally:
            await self._log('end', src)

    async def _log(self, begin_or_end, src):
        progress = self._get_progress(src)
        display = await self.get_source_display(src)
        wl.INFO('wcpan.drive.google') << f'{progress} {begin_or_end} {display}'

    def _get_progress(self, src):
        key = self.get_source_hash(src)
        id_ = self._table[key]
        return f'[{id_}/{self._total}]'

    def _update_counter_table(self, src):
        key = self.get_source_hash(src)
        self._counter += 1
        self._table[key] = self._counter

    async def _get_md5sum(self, local_path):
        loop = asyncio.get_running_loop()
        return await loop.run_in_executor(self._pool, md5sum, local_path)


class UploadQueue(AbstractQueue):

    def __init__(self, drive, jobs):
        super(UploadQueue, self).__init__(drive, jobs)

    async def count_tasks(self, local_path):
        total = 1
        for root, folders, files in os.walk(local_path):
            total = total + len(folders) + len(files)
        return total

    def source_is_folder(self, local_path):
        return op.isdir(local_path)

    async def do_folder(self, local_path, parent_node):
        folder_name = op.basename(local_path)
        node = await self.drive.create_folder(parent_node, folder_name,
                                              exist_ok=True)
        return node

    async def get_children(self, local_path):
        rv = os.listdir(local_path)
        rv = (op.join(local_path, _) for _ in rv)
        return rv

    async def do_file(self, local_path, parent_node):
        node = await upload_from_local(self.drive, parent_node, local_path,
                                       exist_ok=True)
        local_md5 = await self._get_md5sum(local_path)
        if local_md5 != node.md5:
            raise Exception(f'{local_path} md5 mismatch')
        return node

    def get_source_hash(self, local_path):
        return local_path

    async def get_source_display(self, local_path):
        return local_path


class DownloadQueue(AbstractQueue):

    def __init__(self, drive, jobs):
        super(DownloadQueue, self).__init__(drive, jobs)

    async def count_tasks(self, node):
        total = 1
        children = await self.drive.get_children(node)
        count = (self.count_tasks(_) for _ in children)
        count = await asyncio.gather(*count)
        return total + sum(count)

    def source_is_folder(self, node):
        return node.is_folder

    async def do_folder(self, node, local_path):
        full_path = op.join(local_path, node.name)
        os.makedirs(full_path, exist_ok=True)
        return full_path

    async def get_children(self, node):
        return await self.drive.get_children(node)

    async def do_file(self, node, local_path):
        local_path = await download_to_local(self.drive, node, local_path)
        local_md5 = await self._get_md5sum(local_path)
        if local_md5 != node.md5:
            raise Exception(f'{local_path} md5 mismatch')
        return local_path

    def get_source_hash(self, node):
        return node.id_

    async def get_source_display(self, node):
        return await self.drive.get_path(node)


class UploadVerifier(object):

    def __init__(self, drive):
        self._drive = drive
        self._pool = None
        self._raii = None

    async def __aenter__(self):
        async with cl.AsyncExitStack() as stack:
            self._pool = stack.enter_context(cf.ProcessPoolExecutor())
            self._raii = stack.pop_all()
        return self

    async def __aexit__(self, type_, exc, tb):
        await self._raii.aclose()
        self._raii = None
        self._pool = None

    async def run(self, local_path, remote_node):
        if local_path.is_dir():
            await self._run_folder(local_path, remote_node)
        else:
            await self._run_file(local_path, remote_node)

    async def _run_folder(self, local_path, remote_node):
        dir_name = local_path.name

        child_node = await self._get_child_node(local_path, dir_name,
                                                remote_node)
        if not child_node:
            return
        if not child_node.is_folder:
            wl.ERROR('wcpan.drive.google') << f'[NOT_FOLDER] {local_path}'
            return

        wl.INFO('wcpan.drive.google') << f'[OK] {local_path}'

        children = [self.run(child_path, child_node)
                    for child_path in local_path.iterdir()]
        if children:
            await asyncio.wait(children)

    async def _run_file(self, local_path, remote_node):
        file_name = local_path.name
        remote_path = await self._drive.get_path(remote_node)
        remote_path = pl.Path(remote_path, file_name)

        child_node = await self._get_child_node(local_path, file_name,
                                                remote_node)
        if not child_node:
            return
        if not child_node.is_file:
            wl.ERROR('wcpan.drive.google') << f'[NOT_FILE] {local_path}'
            return

        local_md5 = await self._get_md5sum(local_path)
        if local_md5 != child_node.md5:
            wl.ERROR('wcpan.drive.google') << f'[WRONG_MD5] {local_path}'
            return

        wl.INFO('wcpan.drive.google') << f'[OK] {local_path}'

    async def _get_child_node(self, local_path, name, remote_node):
        child_node = await self._drive.get_node_by_name_from_parent(name,
                                                                    remote_node)
        if not child_node:
            wl.ERROR('wcpan.drive.google') << f'[MISSING] {local_path}'
            return None
        if child_node.trashed:
            wl.ERROR('wcpan.drive.google') << f'[TRASHED] {local_path}'
            return None
        return child_node

    async def _get_md5sum(self, local_path):
        loop = asyncio.get_running_loop()
        return await loop.run_in_executor(self._pool, md5sum, local_path)


async def main(args=None):
    if args is None:
        args = sys.argv

    wl.setup((
        'wcpan.drive.google',
    ))

    args = parse_args(args[1:])
    if not args.action:
        await args.fallback_action()
        return 0

    async with Drive(args.conf) as drive:
        return await args.action(drive, args)


def parse_args(args):
    parser = argparse.ArgumentParser('wdg')

    parser.add_argument('-c', '--conf',
        default=get_default_conf_path(),
        help=(
            'specify configuration file path'
            ' (default: %(default)s)'
        ),
    )

    commands = parser.add_subparsers()

    sync_parser = commands.add_parser('sync', aliases=['s'],
        help='synchronize database',
    )
    add_bool_argument(sync_parser, 'verbose', 'v')
    sync_parser.add_argument('-f', '--from', type=int, dest='from_',
        default=None,
        help=(
            'synchronize from certain check point, and do not update cache'
            ' (default: %(default)s)'
        ),
    )
    sync_parser.set_defaults(action=action_sync)

    find_parser = commands.add_parser('find', aliases=['f'],
        help='find files/folders by pattern [offline]',
    )
    add_bool_argument(find_parser, 'id_only')
    add_bool_argument(find_parser, 'include_trash')
    find_parser.add_argument('pattern', type=str)
    find_parser.set_defaults(action=action_find, id_only=False,
                             include_trash=False)

    list_parser = commands.add_parser('list', aliases=['ls'],
        help='list folder [offline]',
    )
    list_parser.set_defaults(action=action_list)
    list_parser.add_argument('id_or_path', type=str)

    tree_parser = commands.add_parser('tree',
        help='recursive list folder [offline]',
    )
    tree_parser.set_defaults(action=action_tree)
    tree_parser.add_argument('id_or_path', type=str)

    dl_parser = commands.add_parser('download', aliases=['dl'],
        help='download files/folders',
    )
    dl_parser.set_defaults(action=action_download)
    dl_parser.add_argument('-j', '--jobs', type=int,
        default=1,
        help=(
            'maximum simultaneously download jobs'
            ' (default: %(default)s)'
        ),
    )
    dl_parser.add_argument('id_or_path', type=str, nargs='+')
    dl_parser.add_argument('destination', type=str)

    ul_parser = commands.add_parser('upload', aliases=['ul'],
        help='upload files/folders',
    )
    ul_parser.set_defaults(action=action_upload)
    ul_parser.add_argument('-j', '--jobs', type=int,
        default=1,
        help=(
            'maximum simultaneously upload jobs'
            ' (default: %(default)s)'
        ),
    )
    ul_parser.add_argument('source', type=str, nargs='+')
    ul_parser.add_argument('id_or_path', type=str)

    rm_parser = commands.add_parser('remove', aliases=['rm'],
        help='trash files/folders',
    )
    rm_parser.set_defaults(action=action_remove)
    rm_parser.add_argument('id_or_path', type=str, nargs='+')

    mv_parser = commands.add_parser('rename', aliases=['mv'],
        help='rename file/folder',
    )
    mv_parser.set_defaults(action=action_rename)
    mv_parser.add_argument('source_id_or_path', type=str)
    mv_parser.add_argument('destination_path', type=str)

    v_parser = commands.add_parser('verify', aliases=['v'],
        help='verify uploaded files/folders',
    )
    v_parser.set_defaults(action=action_verify)
    v_parser.add_argument('source', type=str, nargs='+')
    v_parser.add_argument('id_or_path', type=str)

    d_parser = commands.add_parser('doctor',
        help='check file system error'
    )
    d_parser.set_defaults(action=action_doctor)

    sout = io.StringIO()
    parser.print_help(sout)
    fallback = ft.partial(action_help, sout.getvalue())
    parser.set_defaults(action=None, fallback_action=fallback)

    args = parser.parse_args(args)

    return args


def add_bool_argument(parser, name, short_name=None):
    flag = name.replace('_', '-')
    pos_flags = ['--' + flag]
    if short_name:
        pos_flags.append('-' + short_name)
    neg_flag = '--no-' + flag
    parser.add_argument(*pos_flags, dest=name, action='store_true')
    parser.add_argument(neg_flag, dest=name, action='store_false')


async def action_help(message):
    print(message)


async def action_sync(drive, args):
    chunks = chunks_of(drive.sync(check_point=args.from_), 100)
    async for changes in chunks:
        if not args.verbose:
            print(len(changes))
        else:
            for change in changes:
                print_as_yaml(changes)
    return 0


async def action_find(drive, args):
    nodes = await drive.find_nodes_by_regex(args.pattern)
    if not args.include_trash:
        nodes = (_ for _ in nodes if not _.trashed)
    nodes = (wait_for_value(_.id_, drive.get_path(_)) for _ in nodes)
    nodes = await asyncio.gather(*nodes)
    nodes = dict(nodes)

    if args.id_only:
        for id_ in nodes:
            print(id_)
    else:
        print_id_node_dict(nodes)

    return 0


async def action_list(drive, args):
    node = await get_node_by_id_or_path(drive, args.id_or_path)
    nodes = await drive.get_children(node)
    nodes = {_.id_: _.name for _ in nodes}
    print_id_node_dict(nodes)
    return 0


async def action_tree(drive, args):
    node = await get_node_by_id_or_path(drive, args.id_or_path)
    await traverse_node(drive, node, 0)
    return 0


async def action_download(drive, args):
    node_list = (get_node_by_id_or_path(drive, _) for _ in args.id_or_path)
    node_list = await asyncio.gather(*node_list)
    node_list = [_ for _ in node_list if not _.trashed]

    async with DownloadQueue(drive, args.jobs) as queue_:
        await queue_.run(node_list, args.destination)

    if not queue_.failed:
        return 0
    print('download failed:')
    print_as_yaml(queue_.failed)
    return 1


async def action_upload(drive, args):
    node = await get_node_by_id_or_path(drive, args.id_or_path)

    async with UploadQueue(drive, args.jobs) as queue_:
        await queue_.run(args.source, node)

    if not queue_.failed:
        return 0
    print('upload failed:')
    print_as_yaml(queue_.failed)
    return 1


async def action_remove(drive, args):
    rv = (trash_node(drive, _) for _ in args.id_or_path)
    rv = await asyncio.gather(*rv)
    rv = filter(None, rv)
    rv = list(rv)
    if not rv:
        return 0
    print('trash failed:')
    print_as_yaml(rv)
    return 1


async def action_rename(drive, args):
    node = await get_node_by_id_or_path(drive, args.source_id_or_path)
    node = await drive.rename_node(node, args.destination_path)
    path = await drive.get_path(node)
    return path


async def action_verify(drive, args):
    node = await get_node_by_id_or_path(drive, args.id_or_path)

    async with UploadVerifier(drive) as v:
        tasks = (pl.Path(local_path) for local_path in args.source)
        tasks = [v.run(local_path, node) for local_path in tasks]
        await asyncio.wait(tasks)

    return 0


async def action_doctor(drive, args):
    for node in await drive.find_multiple_parents_nodes():
        print(f'{node.name} has multiple parents, please select one parent:')
        parent_list = (drive.get_node_by_id(_) for _ in node.parent_list)
        parent_list = await asyncio.gather(*parent_list)
        for index, parent_node in enumerate(parent_list):
            print(f'{index}: {parent_node.name}')
        try:
            choice = input()
            choice = int(choice)
            parent = parent_list[choice]
            await drive.set_node_parent_by_id(node, parent.id_)
        except Exception as e:
            print('unknown error, skipped', e)
            continue


async def get_node_by_id_or_path(drive, id_or_path):
    if id_or_path[0] == '/':
        node = await drive.get_node_by_path(id_or_path)
    else:
        node = await drive.get_node_by_id(id_or_path)
    return node


async def traverse_node(drive, node, level):
    if node.is_root:
        print_node('/', level)
    elif level == 0:
        top_path = await drive.get_path(node)
        print_node(top_path, level)
    else:
        print_node(node.name, level)

    if node.is_folder:
        children = await drive.get_children_by_id(node.id_)
        for child in children:
            await traverse_node(drive, child, level + 1)


async def trash_node(drive, id_or_path):
    '''
    :returns: None if succeed, id_or_path if failed
    '''
    node = await get_node_by_id_or_path(drive, id_or_path)
    if not node:
        return id_or_path
    try:
        rv = await drive.trash_node(node)
    except Exception as e:
        wl.EXCEPTION('wcpan.drive.google', e) << 'trash failed'
        return id_or_path
    if not rv:
        return id_or_path
    return None


async def wait_for_value(k, v):
    return k, await v


def md5sum(local_path):
    with open(local_path, 'rb') as fin:
        local_md5 = stream_md5sum(fin)
    return local_md5


async def chunks_of(ag, size):
    chunk = []
    async for item in ag:
        chunk.append(item)
        if len(chunk) == size:
            yield chunk
            chunk = []
    if chunk:
        yield chunk


def print_node(name, level):
    level = ' ' * level
    print(level + name)


def print_as_yaml(data):
    yaml.safe_dump(data, stream=sys.stdout, allow_unicode=True,
                   encoding=sys.stdout.encoding, default_flow_style=False)


def print_id_node_dict(data):
    pairs = sorted(data.items(), key=lambda _: _[1])
    for id_, path in pairs:
        print(f'{id_}: {path}')


sys.exit(asyncio.run(main()))



import asyncio
import contextlib as cl
import enum
import functools as ft
import json
import math
import random
from typing import (Any, AsyncGenerator, AsyncIterator, Callable, Dict,
                    Generator, List, Optional, Text, Tuple, Union)
import urllib.parse as up

import aiohttp
from wcpan.logger import DEBUG, EXCEPTION, INFO, WARNING

from .util import GoogleDriveError, Settings


BACKOFF_FACTOR = 2
BACKOFF_STATUSES = set(('403', '429', '500', '502', '503', '504'))

ContentProducer = Callable[[], AsyncGenerator[bytes, None]]
ReadableContent = Union[bytes, ContentProducer]


def network_timeout(network_method):
    @ft.wraps(network_method)
    async def wrapper(self, *args, **kwargs):
        return await asyncio.wait_for(network_method(self, *args, **kwargs),
                                      self._timeout)
    return wrapper


class Network(object):

    def __init__(self, settings: Settings, timeout: int) -> None:
        self._settings = settings
        self._timeout = timeout
        self._backoff_level = 0
        self._session = None
        self._oauth = None
        self._raii = None

    async def __aenter__(self) -> 'Network':
        async with cl.AsyncExitStack() as stack:
            self._session = await stack.enter_async_context(
                aiohttp.ClientSession())
            self._oauth = await stack.enter_async_context(
                OAuth2Manager(self._session, self._settings))
            self._raii = stack.pop_all()

        return self

    async def __aexit__(self, type_, value, traceback) -> bool:
        await self._raii.aclose()
        self._backoff_level = 0
        self._session = None
        self._oauth = None
        self._raii = None

    @network_timeout
    async def fetch(self,
        method: Text,
        url: Text,
        args: Dict[Text, Any] = None,
        headers: Dict[Text, Text] = None,
        body: ReadableContent = None,
    ) -> 'JSONResponse':
        while True:
            kwargs = await self._prepare_kwargs(method, url, args, headers,
                                                body)

            try:
                response = await self._request_loop(kwargs)
            except aiohttp.ClientConnectionError as e:
                continue

            try:
                return await to_json_response(response)
            except aiohttp.ClientConnectionError as e:
                # The server recived the request, but reading body failed.
                # We are in a broken state, just let client to handle it.
                raise NetworkError() from e

    # NOTE Unlike download, upload cannot set timeout to the whole method.
    # Instead we should set timeout on body callback.
    async def upload(self,
        method: Text,
        url: Text,
        args: Dict[Text, Any] = None,
        headers: Dict[Text, Text] = None,
        body: ReadableContent = None,
    ) -> 'JSONResponse':
        kwargs = await self._prepare_kwargs(method, url, args, headers, body)
        kwargs['timeout'] = 0.0

        try:
            response = await self._request_loop(kwargs)
        except aiohttp.ClientConnectionError as e:
            raise NetworkError() from e

        return await to_json_response(response)

    @network_timeout
    async def download(self,
        method: Text,
        url: Text,
        args: Dict[Text, Any] = None,
        headers: Dict[Text, Text] = None,
        body: ReadableContent = None,
    ) -> 'StreamResponse':
        while True:
            kwargs = await self._prepare_kwargs(method, url, args, headers,
                                                body)
            kwargs['timeout'] = 0.0

            try:
                response = await self._request_loop(kwargs)
            except aiohttp.ClientConnectionError as e:
                continue

            return StreamResponse(response, self._timeout)

    async def _request_loop(self,
        kwargs: Dict[Text, Any],
    ) -> aiohttp.ClientResponse:
        while True:
            await self._wait_backoff()

            try:
                response = await self._session.request(**kwargs)
            except aiohttp.ClientConnectionError as e:
                self._adjust_backoff_level(True)
                raise

            status = str(response.status)
            rv = await self._check_status(status, response)
            if rv == Status.OK:
                return response
            if rv == Status.REFRESH:
                await self._oauth.renew_token()
                await self._update_token_header(kwargs['headers'])
                continue
            if rv == Status.BACKOFF:
                continue

            json_ = await response.json()
            raise ResponseError(status, response, json_)

    async def _prepare_kwargs(self,
        method: Text,
        url: Text,
        args: Optional[Dict[Text, Any]],
        headers: Optional[Dict[Text, Text]],
        body: Optional[ReadableContent],
    ) -> Dict[Text, Any]:
        kwargs = {
            'method': method,
            'url': url,
            'headers': await self._prepare_headers(headers),
        }
        if args is not None:
            kwargs['params'] = list(normalize_query_string(args))
        if body is not None:
            kwargs['data'] = body if not callable(body) else body()
        return kwargs

    async def _prepare_headers(self,
        headers: Optional[Dict[Text, Text]],
    ) -> Dict[Text, Text]:
        if headers is None:
            h = {}
        else:
            h = {k: v if isinstance(v, (bytes, str)) or v is None else str(v)
                 for k, v in headers.items()}
        await self._update_token_header(h)
        return h

    async def _update_token_header(self, headers: Dict[Text, Text]) -> None:
        token = await self._oauth.get_access_token()
        headers['Authorization'] = f'Bearer {token}'

    async def _check_status(self,
        status: Text,
        response: aiohttp.ClientResponse,
    ) -> 'Status':
        backoff = await backoff_needed(status, response)
        self._adjust_backoff_level(backoff)
        if backoff:
            # rate limit error, too many request, server error
            return Status.BACKOFF

        # normal response
        if status[0] in ('1', '2', '3'):
            return Status.OK

        # need to refresh access token
        if status == '401':
            return Status.REFRESH

        # otherwise it is an error
        return Status.UNKNOWN

    def _adjust_backoff_level(self, backoff: bool) -> None:
        if backoff:
            self._backoff_level = min(self._backoff_level + 2, 10)
        else:
            self._backoff_level = max(self._backoff_level - 1, 0)

    async def _wait_backoff(self) -> None:
        if self._backoff_level <= 0:
            return
        seed = random.random()
        power = 2 ** self._backoff_level
        s_delay = math.floor(seed * power * BACKOFF_FACTOR)
        s_delay = min(100, s_delay)
        DEBUG('wcpan.drive.google') << 'backoff for' << s_delay
        await asyncio.sleep(s_delay)


class Request(object):

    def __init__(self, request: aiohttp.RequestInfo) -> None:
        self._request = request

    @property
    def uri(self) -> Text:
        return self._request.url

    @property
    def method(self) -> Text:
        return self._request.method

    @property
    def headers(self) -> Dict[Text, Text]:
        return self._request.headers


class Response(object):

    def __init__(self, response: aiohttp.ClientResponse) -> None:
        self._response = response
        self._status = str(response.status)
        self._request = Request(response.request_info)

    @property
    def status(self):
        return self._status

    def get_header(self, key: Text) -> Text:
        h = self._response.headers.getall(key)
        return None if not h else h[0]


class JSONResponse(Response):

    def __init__(self,
        response: aiohttp.ClientResponse,
        json_: Dict[Text, Any],
    ) -> None:
        super().__init__(response)
        self._json = json_

    @property
    def json(self) -> Dict[Text, Any]:
        return self._json


class StreamResponse(Response):

    def __init__(self, response: aiohttp.ClientResponse, timeout: int) -> None:
        super().__init__(response)
        self._timeout = timeout

    async def __aenter__(self) -> 'StreamResponse':
        await self._response.__aenter__()
        return self

    async def __aexit__(self, type_, exc, tb) -> bool:
        await self._response.__aexit__(type_, exc, tb)

    async def chunks(self) -> AsyncIterator[bytes]:
        g = self._response.content.iter_any()
        while True:
            try:
                v = await asyncio.wait_for(g.__anext__(), self._timeout)
            except StopAsyncIteration:
                break
            yield v

    async def read(self, length):
        return await self._response.content.read(length)


async def to_json_response(response: aiohttp.ClientResponse) -> JSONResponse:
    async with response:
        if response.content_type == 'application/json':
            json_ = await response.json()
        else:
            json_ = await response.text()
    return JSONResponse(response, json_)


class ResponseError(GoogleDriveError):

    def __init__(self,
        status: Text,
        response: aiohttp.ClientResponse,
        json_: Dict[Text, Any],
    ) -> None:
        self._status = status
        self._response = response
        self._message = f'{self.status} {self._response.reason} - {json_}'
        self._json = json_

    def __str__(self) -> Text:
        return self._message

    @property
    def status(self) -> Text:
        return self._status

    @property
    def json(self) -> Any:
        return self._json


class NetworkError(GoogleDriveError):
    pass


class AuthenticationError(GoogleDriveError):
    pass


class OAuth2Manager(object):

    def __init__(self,
        session: aiohttp.ClientSession,
        settings: Settings,
    ) -> None:
        self._session = session
        self._settings = settings
        self._lock = asyncio.Condition()
        self._refreshing = False
        self._error = False
        self._oauth = None
        self._raii = None

    async def __aenter__(self) -> 'OAuth2Manager':
        oauth2_info = await self._settings.load_oauth2_info()

        async with cl.AsyncExitStack() as stack:
            self._oauth = await stack.enter_async_context(
                CommandLineGoogleDriveOAuth2(
                    self._session,
                    oauth2_info['client_id'],
                    oauth2_info['client_secret'],
                    oauth2_info['redirect_uri'],
                    oauth2_info['access_token'],
                    oauth2_info['refresh_token'],
                ))
            await self._settings.save_oauth2_info(self._oauth.access_token,
                                                  self._oauth.refresh_token)
            self._raii = stack.pop_all()

        return self

    async def __aexit__(self, type_, exc, tb) -> bool:
        await self._raii.aclose()
        self._raii = None
        self._oauth = None
        self._error = False
        self._refreshing = False

    async def get_access_token(self) -> Text:
        if self._refreshing:
            async with self._lock:
                await self._lock.wait()
        if self._error:
            raise AuthenticationError()
        return self._oauth.access_token

    async def renew_token(self):
        if self._refreshing:
            async with self._lock:
                await self._lock.wait()
            return

        async with self._guard():
            try:
                await self._oauth.refresh()
                await self._settings.save_oauth2_info(self._oauth.access_token,
                                                      self._oauth.refresh_token)
            except Exception as e:
                EXCEPTION('wcpan.drive.google', e) << 'error on refresh token'
                self._error = True
                raise
            self._error = False

        DEBUG('wcpan.drive.google') << 'refresh access token'

    @cl.asynccontextmanager
    async def _guard(self):
        self._refreshing = True
        try:
            yield
        finally:
            self._refreshing = False
            async with self._lock:
                self._lock.notify_all()


class CommandLineGoogleDriveOAuth2(object):

    def __init__(self,
        session: aiohttp.ClientSession,
        client_id: Text,
        client_secret: Text,
        redirect_uri: Text,
        access_token: Text = None,
        refresh_token: Text = None
    ) -> None:
        self._session = session
        self._client_id = client_id
        self._client_secret = client_secret
        self._redirect_uri = redirect_uri
        self._access_token = access_token
        self._refresh_token = refresh_token

    async def __aenter__(self) -> 'CommandLineGoogleDriveOAuth2':
        if self._access_token is None:
            await self._fetch_access_token()
        return self

    async def __aexit__(self, type_, value, traceback) -> bool:
        pass

    @property
    def access_token(self) -> Text:
        assert self._access_token is not None
        return self._access_token

    @property
    def refresh_token(self) -> Union[Text, None]:
        return self._refresh_token

    async def refresh(self) -> None:
        headers = {
            'Content-Type': 'application/x-www-form-urlencoded',
        }
        body = up.urlencode({
            'client_id': self._client_id,
            'client_secret': self._client_secret,
            'refresh_token': self._refresh_token,
            'grant_type': 'refresh_token',
        })

        async with self._session.post(self.oauth_access_token_url,
                                      headers=headers, data=body) as response:
            response.raise_for_status()
            token = await response.json()
        self._save_token(token)

    def _save_token(self, token: Dict[Text, Any]) -> None:
        self._access_token = token['access_token']
        if 'refresh_token' in token:
            self._refresh_token = token['refresh_token']

    async def _fetch_access_token(self) -> None:
        # get code on success
        code = await self._authorize_redirect()
        token = await self._get_authenticated_user(code=code)
        self._save_token(token)

    async def _authorize_redirect(self) -> Text:
        kwargs = {
            'redirect_uri': self._redirect_uri,
            'client_id': self._client_id,
            'response_type': 'code',
            'scope': ' '.join(self.scopes),
        }

        url = up.urlparse(self.oauth_authorize_url)
        url = up.urlunparse((
            url[0],
            url[1],
            url[2],
            url[3],
            up.urlencode(kwargs),
            url[5],
        ))
        return await self.redirect(url)

    # NOTE Google only
    @property
    def oauth_authorize_url(self) -> Text:
        return 'https://accounts.google.com/o/oauth2/auth'

    # NOTE Google only
    @property
    def oauth_access_token_url(self) -> Text:
        return 'https://accounts.google.com/o/oauth2/token'

    # NOTE Google only
    @property
    def scopes(self) -> List[Text]:
        return [
            'https://www.googleapis.com/auth/drive',
        ]

    # NOTE Google only?
    async def _get_authenticated_user(self, code: Text) -> Dict[Text, Any]:
        headers = {'Content-Type': 'application/x-www-form-urlencoded'}
        body = up.urlencode({
            'redirect_uri': self._redirect_uri,
            'code': code,
            'client_id': self._client_id,
            'client_secret': self._client_secret,
            'grant_type': 'authorization_code',
        })
        async with self._session.post(self.oauth_access_token_url,
                                      headers=headers, data=body) as response:
            response.raise_for_status()
            return await response.json()

    # NOTE Use case depends
    async def redirect(self, url: Text) -> Text:
        print(url)
        return input().strip()


class Status(enum.Enum):

    OK = enum.auto()
    REFRESH = enum.auto()
    BACKOFF = enum.auto()
    UNKNOWN = enum.auto()


async def backoff_needed(
    status: Text,
    response: aiohttp.ClientResponse,
) -> bool:
    if status not in BACKOFF_STATUSES:
        return False

    # not all 403 errors are rate limit error
    if status == '403':
        msg = await response.json()
        if not msg:
            # undefined behavior, probably a server problem, better backoff
            WARNING('wcpan.drive.google') << '403 with empty error message'
            return True
        domain = msg['error']['errors'][0]['domain']
        if domain != 'usageLimits':
            return False

    return True


def normalize_query_string(
    qs: Dict[Text, Any],
) -> Generator[Tuple[Text, Text], None, None]:
    for key, value in qs.items():
        if isinstance(value, bool):
            value = 'true' if value else 'false'
        elif isinstance(value, (int, float)):
            value = str(value)
        elif not isinstance(value, str):
            raise ValueError('unknown type in query string')
        yield key, value



import asyncio
import contextlib as cl
import functools as ft
import mimetypes
import os
import os.path as op
import re
from typing import (Any, AsyncGenerator, Awaitable, Dict, Generator, List,
                    Optional, Text, Tuple, Union)

from wcpan.logger import INFO, WARNING, DEBUG, EXCEPTION

from .api import Client
from .cache import Cache, Node, node_from_api, dict_from_node
from .network import ContentProducer, ResponseError, Response, NetworkError
from .util import Settings, GoogleDriveError, CHUNK_SIZE


FILE_FIELDS = ','.join([
    'id',
    'name',
    'mimeType',
    'trashed',
    'parents',
    'createdTime',
    'modifiedTime',
    'md5Checksum',
    'size',
    'shared',
    'ownedByMe',
    'imageMediaMetadata',
    'videoMediaMetadata',
])
CHANGE_FIELDS = ','.join([
    'nextPageToken',
    'newStartPageToken',
    f'changes(fileId,removed,file({FILE_FIELDS}))',
])


class Drive(object):

    def __init__(self, conf_path: Text = None, timeout: int = 60) -> None:
        self._settings = Settings(conf_path)
        self._timeout = timeout
        self._client = None
        self._db = None
        self._sync_lock = asyncio.Lock()
        self._raii = None

    async def __aenter__(self) -> 'Drive':
        async with cl.AsyncExitStack() as stack:
            self._client = await stack.enter_async_context(
                Client(self._settings, self._timeout))
            dsn = self._settings['nodes_database_file']
            self._db = await stack.enter_async_context(Cache(dsn))
            self._raii = stack.pop_all()
        return self

    async def __aexit__(self, exc_type, exc, tb) -> bool:
        await self._raii.aclose()
        self._client = None
        self._db = None
        self._raii = None

    async def sync(self,
        check_point: int = None,
    ) -> AsyncGenerator[Dict[Text, Any], None]:
        async with self._sync_lock:
            async for change in self._real_sync(check_point):
                yield change

    async def get_root_node(self) -> Node:
        return await self._db.get_root_node()

    async def get_node_by_id(self, node_id: Text) -> Node:
        return await self._db.get_node_by_id(node_id)

    async def get_node_by_path(self, path: Text) -> Node:
        return await self._db.get_node_by_path(path)

    async def get_path(self, node: Node) -> Text:
        return await self._db.get_path_by_id(node.id_)

    async def get_path_by_id(self, node_id: Text) -> Text:
        return await self._db.get_path_by_id(node_id)

    async def get_node_by_name_from_parent_id(self,
        name: Text,
        parent_id: Text,
    ) -> Node:
        return await self._db.get_node_by_name_from_parent_id(name, parent_id)

    async def get_node_by_name_from_parent(self,
        name: Text,
        parent: Node,
    ) -> Node:
        return await self._db.get_node_by_name_from_parent_id(name, parent.id_)

    async def get_children(self, node: Node) -> List[Node]:
        return await self._db.get_children_by_id(node.id_)

    async def get_children_by_id(self, node_id: Text) -> List[Node]:
        return await self._db.get_children_by_id(node_id)

    async def find_nodes_by_regex(self, pattern: Text) -> List[Node]:
        return await self._db.find_nodes_by_regex(pattern)

    async def find_duplicate_nodes(self) -> List[Node]:
        return await self._db.find_duplicate_nodes()

    async def find_orphan_nodes(self) -> List[Node]:
        return await self._db.find_orphan_nodes()

    async def find_multiple_parents_nodes(self) -> List[Node]:
        return await self._db.find_multiple_parents_nodes()

    async def download_by_id(self, node_id: Text) -> 'ReadableFile':
        node = await self.get_node_by_id(node_id)
        return await self.download(node)

    async def download(self, node: Node) -> 'ReadableFile':
        # sanity check
        if not node:
            raise ValueError('node is none')
        if node.is_folder:
            raise ValueError('node should be a file')

        rf = ReadableFile(self._client.files.download, node)
        return rf

    async def create_folder(self,
        parent_node: Node,
        folder_name: Text,
        exist_ok: bool = False,
    ) -> Node:
        # sanity check
        if not parent_node:
            raise UploadError('invalid parent node')
        if not parent_node.is_folder:
            raise UploadError('invalid parent node')
        if not folder_name:
            raise UploadError('invalid folder name')

        # do not create again if there is a same file
        node = await self.fetch_node_by_name_from_parent_id(folder_name,
                                                            parent_node.id_)
        if node:
            if exist_ok:
                INFO('wcpan.drive.google') << 'skipped (existing)' << folder_name
                return node
            else:
                raise FileConflictedError(node)

        api = self._client.files
        rv = await api.create_folder(folder_name=folder_name,
                                     parent_id=parent_node.id_)
        rv = rv.json
        node = await self.fetch_node_by_id(rv['id'])

        return node

    async def upload_by_id(self,
        parent_id: Text,
        file_name: Text,
        file_size: int = None,
        mime_type: Text = None,
    ) -> 'WritableFile':
        node = await self.get_node_by_id(parent_id)
        return await self.upload(node, file_name, file_size, mime_type)

    async def upload(self,
        parent_node: Node,
        file_name: Text,
        file_size: int = None,
        mime_type: Text = None,
    ) -> 'WritableFile':
        # sanity check
        if not parent_node:
            raise UploadError('invalid parent node')
        if not parent_node.is_folder:
            raise UploadError('invalid parent node')
        if not file_name:
            raise UploadError('invalid file name')

        # do not upload if remote exists a same file
        node = await self.fetch_node_by_name_from_parent_id(file_name,
                                                            parent_node.id_)
        if node:
            raise FileConflictedError(node)

        api = self._client.files
        wf = WritableFile(initiate=api.initiate_uploading,
                          upload=api.upload,
                          get_status=api.get_upload_status,
                          touch=api.create_empty_file,
                          parent_id=parent_node.id_,
                          name=file_name,
                          timeout=self._timeout,
                          size=file_size,
                          mime_type=mime_type)
        return wf

    async def fetch_node_by_name_from_parent_id(self,
        name: Text,
        parent_id: Text,
    ) -> Node:
        safe_name = re.sub(r"[\\']", r"\\\g<0>", name)
        query = f"'{parent_id}' in parents and name = '{safe_name}'"
        fields = f'files({FILE_FIELDS})'
        try:
            rv = await self._client.files.list_(q=query, fields=fields)
        except ResponseError as e:
            if e.status == '400':
                DEBUG('wcpan.drive.google') << 'invalid query string:' << query
                raise InvalidNameError(name) from e
            if e.status == '404':
                raise ParentNotFoundError(parent_id) from e
            raise

        rv = rv.json
        files = rv['files']
        if not files:
            return None

        node = node_from_api(files[0])
        return node

    async def fetch_node_by_id(self, node_id: Text) -> Node:
        try:
            rv = await self._client.files.get(node_id, fields=FILE_FIELDS)
        except ResponseError as e:
            if e.status == '404':
                raise NodeNotFoundError(node_id) from e
        rv = rv.json
        node = node_from_api(rv)
        return node

    async def trash_node_by_id(self, node_id: Text) -> Node:
        node = await self.get_root_node()
        if node_id == node.id_:
            return
        await self._client.files.update(node_id, trashed=True)

        node = await self.get_node_by_id(node_id)
        node.trashed = True
        return node

    async def trash_node(self, node: Node) -> Node:
        return await self.trash_node_by_id(node.id_)

    async def rename_node_by_path(self, src_path: Text, dst_path: Text) -> Node:
        node = await self.get_node_by_path(src_path)
        if not node:
            raise NodeNotFoundError(src_path)
        return await self.rename_node(node, dst_path)

    async def rename_node_by_id(self, node_id: Text, dst_path: Text) -> Node:
        node = await self.get_node_by_id(node_id)
        return await self.rename_node(node, dst_path)

    async def rename_node(self, src_node: Node, dst_path: Text) -> Node:
        '''
        Rename or move `src_node` to `dst_path`. `dst_path` can be a file name
        or an absolute path.
        If `dst_path` is a file and already exists, `FileConflictedError` will
        be raised.
        If `dst_path` is a folder, `src_node` will be moved to there without
        renaming.
        If `dst_path` does not exist yet, `src_node` will be moved and rename to
        `dst_path`.
        '''
        # sanity check
        if not src_node:
            raise ValueError('source node is none')

        parent, dst_name = await self._get_dst_info(dst_path)
        await self._inner_rename_node(src_node, parent, dst_name)

        node = await self.fetch_node_by_id(src_node.id_)
        return node

    async def set_node_parent_by_id(self, node: Node, parent_id: Text) -> None:
        remove_parents = [_ for _ in node.parent_list if _ != parent_id]
        api = self._client.files
        await api.update(node.id_, remove_parents=remove_parents)


    async def walk(self,
        node: Node,
    ) -> AsyncGenerator[Tuple[Node, List[Node], List[Node]], None]:
        if not node.is_folder:
            return
        q = [node]
        while q:
            node = q[0]
            del q[0]
            children = await self.get_children(node)
            folders = list(filter(lambda _: _.is_folder, children))
            files = list(filter(lambda _: _.is_file, children))
            yield node, folders, files
            q.extend(folders)

    async def _real_sync(self,
        check_point: int,
    ) -> AsyncGenerator[Dict[Text, Any], None]:
        dry_run = check_point is not None
        if dry_run and check_point > 0:
            check_point = str(check_point)
        else:
            try:
                check_point = await self._db.get_metadata('check_point')
            except KeyError:
                check_point = '1'

        # first time, get root node
        if not dry_run and check_point == '1':
            rv = await self._client.files.get('root', fields=FILE_FIELDS)
            rv = rv.json
            rv['name'] = None
            rv['parents'] = []
            node = node_from_api(rv)
            await self._db.insert_node(node)

        new_start_page_token = None
        changes_list_args = {
            'page_token': check_point,
            'page_size': 1000,
            'restrict_to_my_drive': True,
            'fields': CHANGE_FIELDS,
        }

        while new_start_page_token is None:
            rv = await self._client.changes.list_(**changes_list_args)
            rv = rv.json
            next_page_token = rv.get('nextPageToken', None)
            new_start_page_token = rv.get('newStartPageToken', None)
            changes = rv['changes']

            check_point = next_page_token if next_page_token is not None else new_start_page_token

            if not dry_run:
                await self._db.apply_changes(changes, check_point)

            changes_list_args['page_token'] = check_point

            for change in transform_changes(changes):
                yield change

    async def _get_dst_info(self, dst_path: Text) -> Tuple[Node, Text]:
        if not op.isabs(dst_path):
            if op.basename(dst_path) != dst_path:
                raise ValueError(f'invalid path: {dst_path}')
            # rename only
            return None, dst_path
        else:
            dst_node = await self.get_node_by_path(dst_path)
            if not dst_node:
                # move to the parent folder
                dst_folder, dst_name = op.split(dst_path)
                parent = await self.get_node_by_path(dst_folder)
                return parent, dst_name
            if dst_node.is_file:
                # do not overwrite existing file
                raise FileConflictedError(dst_path)
            # just move to this folder
            return dst_node, None

    async def _inner_rename_node(self,
        node: Node,
        new_parent: Optional[Node],
        name: Optional[Text],
    ) -> Response:
        # sanity check
        if not new_parent and not name:
            raise ValueError('invalid arguments')
        fnbnfpi = self.fetch_node_by_name_from_parent_id
        new_parent_id = node.parent_id if not new_parent else new_parent.id_
        new_name = node.name if not name else name
        new_node = await fnbnfpi(new_name, new_parent_id)
        if new_node:
            raise FileConflictedError(new_node)

        kwargs = {
            'file_id': node.id_,
        }
        if name:
            kwargs['name'] = name
        if new_parent and new_parent.id_ != node.parent_id:
            kwargs['add_parents'] = [new_parent.id_]
            kwargs['remove_parents'] = [node.parent_id]

        rv = await self._client.files.update(**kwargs)
        return rv


class ReadableFile(object):

    def __init__(self, download: Any, node: Node) -> None:
        self._download = download
        self._node = node
        self._offset = 0
        self._response = None
        self._rsps = None

    async def __aenter__(self) -> 'ReadableFile':
        return self

    async def __aexit__(self, type_, exc, tb) -> bool:
        await self._close_response()

    async def __aiter__(self) -> bytes:
        await self._open_response()
        async for chunk in self._response.chunks():
            yield chunk

    async def read(self, length: int) -> bytes:
        await self._open_response()
        return await self._response.read(length)

    async def seek(self, offset: int) -> None:
        self._offset = offset
        await self._close_response()
        await self._open_response()

    async def _download_from_offset(self) -> 'aiohttp.StreamResponse':
        return await self._download(file_id=self._node.id_,
                                    range_=(self._offset, self._node.size))

    async def _open_response(self) -> None:
        if not self._response:
            async with cl.AsyncExitStack() as stack:
                self._response = await stack.enter_async_context(
                    await self._download_from_offset())
                self._rsps = stack.pop_all()

    async def _close_response(self) -> None:
        if self._response:
            await self._rsps.aclose()
            self._response = None
            self._rsps = None


class WritableFile(object):

    def __init__(self,
        initiate: Any,
        upload: Any,
        get_status: Any,
        touch: Any,
        parent_id: Text,
        name: Text,
        timeout: float,
        size: int = None,
        mime_type: Text = None,
    ) -> None:
        self._initiate = initiate
        self._upload = upload
        self._get_status = get_status
        self._touch = touch
        self._parent_id = parent_id
        self._name = name
        self._timeout = timeout
        self._size = size
        self._mime_type = mime_type
        self._url = None
        self._offset = None
        self._queue = asyncio.Queue(maxsize=1)
        self._bg = None
        self._rv = None

    async def __aenter__(self) -> 'WritableFile':
        if self._size > 0:
            self._url = await self._get_session_url()
            self._bg = None
        else:
            self._url = None
            self._bg = self._touch_empty()
        self._offset = 0
        self._queue = asyncio.Queue(maxsize=1)
        self._rv = None

        return self

    async def __aexit__(self, type_, exc, tb) -> bool:
        rv = await self._bg
        self._rv = rv.json
        self._bg = None
        self._queue = None
        self._offset = None
        self._url = None

    async def tell(self) -> int:
        return await self._get_offset()

    async def seek(self, offset: int) -> None:
        self._offset = offset
        await self._close_request()
        await self._open_request()

    async def write(self, chunk: bytes) -> int:
        await self._open_request()
        feed = self._queue.put(chunk)
        feed = asyncio.wait_for(feed, timeout=self._timeout)
        feed = asyncio.create_task(feed)
        await asyncio.wait([feed, self._bg],
                           return_when=asyncio.FIRST_COMPLETED)
        if not feed.done():
            # background connection ended
            feed.cancel()
        # consume exception
        feed.result()
        return len(chunk)

    @property
    def id_(self) -> Text:
        return self._rv['id']

    async def _close_request(self) -> None:
        if not self._bg:
            return None
        if not self._bg.done():
            self._bg.cancel()
        try:
            rv = await self._bg
            return rv
        except (Exception, asyncio.CancelledError) as e:
            EXCEPTION('wcpan.drive.google', e) << 'close'
        finally:
            self._queue = asyncio.Queue(maxsize=1)
            self._bg = None

    async def _open_request(self) -> None:
        if self._bg:
            return
        f = self._upload_to()
        loop = asyncio.get_running_loop()
        self._bg = loop.create_task(f)

    async def _produce(self) -> AsyncGenerator[bytes, None]:
        while True:
            async with self._get_one_chunk() as chunk:
                if not chunk:
                    break
                yield chunk

    @cl.asynccontextmanager
    async def _get_one_chunk(self) -> AsyncGenerator[bytes, None]:
        chunk = await self._queue.get()
        try:
            yield chunk
        finally:
            self._queue.task_done()

    async def _get_session_url(self) -> Text:
        rv = await self._initiate(file_name=self._name,
                                  total_file_size=self._size,
                                  parent_id=self._parent_id,
                                  mime_type=self._mime_type)
        url = rv.get_header('Location')
        return url

    async def _upload_to(self) -> 'aiohttp.ClientResponse':
        try:
            rv = await self._upload(self._url, producer=self._produce,
                                    offset=self._offset,
                                    total_file_size=self._size,
                                    mime_type=self._mime_type)
        except ResponseError as e:
            if e.status == '404':
                raise UploadError('the upload session has been expired') from e
            raise
        return rv

    async def _get_offset(self) -> int:
        try:
            rv = await self._get_status(self._url, self._size)
        except ResponseError as e:
            if e.status == '410':
                # This means the temporary URL has been cleaned up by Google
                # Drive, so the client has to start over again.
                raise UploadError('the upload session has been expired') from e
            raise

        if rv.status != '308':
            raise UploadError(f'invalid upload status: {rv.status}')
        
        try:
            rv = rv.get_header('Range')
        except KeyError:
            # no data yet
            return 0

        rv = re.match(r'bytes=0-(\d+)', rv)
        if not rv:
            raise UploadError('invalid upload range')
        rv = int(rv.group(1))
        return rv

    async def _touch_empty(self) -> 'aiohttp.ClientResponse':
        rv = await self._touch(self._name, self._parent_id, self._mime_type)
        return rv


class DownloadError(GoogleDriveError):

    def __init__(self, message: Text) -> None:
        self._message = message

    def __str__(self) -> Text:
        return self._message


class UploadError(GoogleDriveError):

    def __init__(self, message: Text) -> None:
        self._message = message

    def __str__(self) -> Text:
        return self._message


class RemoteNodeError(GoogleDriveError):

    pass


class FileConflictedError(RemoteNodeError):

    def __init__(self, node: Node) -> None:
        self._node = node

    def __str__(self) -> Text:
        return f'remote file already exists: {self._node.name}'

    @property
    def node(self) -> Node:
        return self._node


class InvalidNameError(GoogleDriveError):

    def __init__(self, name: Text) -> None:
        self._name = name

    def __str__(self) -> Text:
        return f'invalid name: {self._name}'


class ParentNotFoundError(RemoteNodeError):

    def __init__(self, id_: Text) -> None:
        self._id = id_

    def __str__(self) -> Text:
        return f'remote parent id not found: {self._id}'


class NodeNotFoundError(RemoteNodeError):

    def __init__(self, path_or_id: Text) -> None:
        self._path_or_id = path_or_id

    def __str__(self) -> Text:
        return f'remote node not found: {self._path_or_id}'


async def download_to_local_by_id(
    drive: Drive,
    node_id: Text,
    path: Text,
) -> Text:
    node = await drive.get_node_by_id(node_id)
    return await download_to_local(drive, node, path)


async def download_to_local(drive: Drive, node: Node, path: Text) -> Text:
    if not op.isdir(path):
        raise ValueError(f'{path} does not exist')

    # check if exists
    complete_path = op.join(path, node.name)
    if op.isfile(complete_path):
        return complete_path

    # exists but not a file
    if op.exists(complete_path):
        raise DownloadError(f'{complete_path} exists but is not a file')

    # if the file is empty, no need to download
    if node.size <= 0:
        open(complete_path, 'w').close()
        return complete_path

    # resume download
    tmp_path = complete_path + '.__tmp__'
    if op.isfile(tmp_path):
        offset = op.getsize(tmp_path)
        if offset > node.size:
            raise DownloadError(
                f'local file size of `{complete_path}` is greater then remote'
                f' ({offset} > {node.size})')
    elif op.exists(tmp_path):
        raise DownloadError(f'{complete_path} exists but is not a file')
    else:
        offset = 0

    if offset < node.size:
        async with await drive.download(node) as fin:
            await fin.seek(offset)
            with open(tmp_path, 'ab') as fout:
                while True:
                    try:
                        async for chunk in fin:
                            fout.write(chunk)
                        break
                    except NetworkError as e:
                        EXCEPTION('wcpan.drive.google', e) << 'download'

                    offset = fout.tell()
                    await fin.seek(offset)

    # rename it back if completed
    os.rename(tmp_path, complete_path)

    return complete_path


async def upload_from_local_by_id(
    drive: Drive,
    parent_id: Text,
    file_path: Text,
    exist_ok: bool = False,
) -> Node:
    node = await drive.get_node_by_id(parent_id)
    return await upload_from_local(drive, node, file_path, exist_ok)


async def upload_from_local(
    drive: Drive,
    parent_node: Node,
    file_path: Text,
    exist_ok: bool = False,
) -> Node:
    # sanity check
    if not op.isfile(file_path):
        raise UploadError('invalid file path')

    file_name = op.basename(file_path)
    total_file_size = op.getsize(file_path)
    mt, _ = mimetypes.guess_type(file_path)

    try:
        fout = await drive.upload(parent_node=parent_node,
                                  file_name=file_name,
                                  file_size=total_file_size,
                                  mime_type=mt)
    except FileConflictedError as e:
        if not exist_ok:
            raise
        return e.node

    async with fout:
        with open(file_path, 'rb') as fin:
            while True:
                try:
                    await upload_feed(fin, fout)
                    break
                except UploadError as e:
                    raise
                except Exception as e:
                    EXCEPTION('wcpan.drive.google', e) << 'upload feed'

                await upload_continue(fin, fout)

    node = await drive.fetch_node_by_id(fout.id_)
    return node


async def upload_feed(fin, fout) -> None:
    while True:
        chunk = fin.read(CHUNK_SIZE)
        if not chunk:
            break
        await fout.write(chunk)


async def upload_continue(fin, fout) -> None:
    offset = await fout.tell()
    await fout.seek(offset)
    fin.seek(offset, os.SEEK_SET)


def transform_changes(
    change_list: List[Dict[Text, Any]],
) -> Generator[Dict[Text, Any], None, None]:
    for change in change_list:
        is_removed = change['removed']
        if is_removed:
            yield {
                'removed': True,
                'id': change['fileId'],
            }
            continue

        file_ = change['file']
        is_shared = file_['shared']
        is_owned_by_me = file_['ownedByMe']
        if is_shared or not is_owned_by_me:
            continue

        file_ = node_from_api(file_)
        file_ = dict_from_node(file_)
        yield {
            'removed': False,
            'node': file_,
        }



import hashlib
import json
import os
import os.path as op
from typing import Any, BinaryIO, Dict, Text

import yaml


CHUNK_SIZE = 64 * 1024
FOLDER_MIME_TYPE = 'application/vnd.google-apps.folder'


class GoogleDriveError(Exception):
    pass


class Settings(object):

    def __init__(self, path: Text = None) -> None:
        self._path = find_conf_path(path)
        self._data = {
            'version': 1,
            'save_credentials': False,
            'client_config_backend': 'settings',
            'nodes_database_file': ':memory:',
        }
        self._initialize()

    def _initialize(self) -> None:
        # default values for file
        self._data['version'] = 1
        self._data['client_config_backend'] = 'file'
        self._data['client_config_file'] = op.join(self._path, 'client_secret.json')
        self._data['get_refresh_token'] = True
        self._data['save_credentials'] = True
        self._data['save_credentials_backend'] = 'file'
        self._data['save_credentials_file'] = op.join(self._path, 'oauth_token.yaml')
        self._data['nodes_database_file'] = op.join(self._path, 'nodes.db')

        os.makedirs(self._path, exist_ok=True)
        path = op.join(self._path, 'settings.yaml')
        if not op.exists(path):
            rv = yaml.dump(self._data, default_flow_style=False)
            with open(path, 'w') as fout:
                fout.write(rv)
        else:
            with open(path, 'r') as fin:
                rv = yaml.load(fin)
            self._data.update(rv)

    def __getitem__(self, key: Text) -> Any:
        return self._data[key]

    async def load_oauth2_info(self) -> Dict[Text, Any]:
        if 'client_config_file' not in self._data:
            raise ValueError('`client_config_file` not found')

        # load API key
        with open(self._data['client_config_file'], 'r') as fin:
            client = json.load(fin)
        if 'installed' not in client:
            raise ValueError('credential should be an installed application')
        client = client['installed']
        redirect_uri = client['redirect_uris'][0]
        client_id = client['client_id']
        client_secret = client['client_secret']

        # load refresh token
        token_path = self._data['save_credentials_file']
        if not op.isfile(token_path):
            access_token = None
            refresh_token = None
        else:
            with open(token_path, 'r') as fin:
                token = yaml.safe_load(fin)
            if token.get('version', 0) != 1:
                raise ValueError('wrong token file')
            access_token = token['access_token']
            refresh_token = token['refresh_token']

        return {
            'client_id': client_id,
            'client_secret': client_secret,
            'redirect_uri': redirect_uri,
            'access_token': access_token,
            'refresh_token': refresh_token,
        }

    async def save_oauth2_info(self,
            access_token: Text,
            refresh_token: Text,
        ) -> None:
        token = {
            'version': 1,
            'access_token': access_token,
            'refresh_token': refresh_token,
        }
        token_path = self._data['save_credentials_file']

        # save refresh token
        rv = yaml.dump(token, default_flow_style=False)
        with open(token_path, 'w') as fout:
            fout.write(rv)


def stream_md5sum(input_stream: BinaryIO) -> Text:
    hasher = hashlib.md5()
    while True:
        chunk = input_stream.read(CHUNK_SIZE)
        if not chunk:
            break
        hasher.update(chunk)
    return hasher.hexdigest()


def find_conf_path(path: Text = None):
    if path is not None:
        return path
    return get_default_conf_path()


def get_default_conf_path():
    path = '~/.cache/wcpan/drive/google'
    path = op.expanduser(path)
    return path



import asyncio
import contextlib as cl
import concurrent.futures as cf
import functools as ft
import pathlib as pl
import re
import sqlite3
from typing import Any, Dict, List, Optional, Text, Union

import arrow

from . import util as u


SQL_CREATE_TABLES = [
    '''
    CREATE TABLE metadata (
        key TEXT NOT NULL,
        value TEXT,
        PRIMARY KEY (key)
    );
    ''',

    '''
    CREATE TABLE nodes (
        id TEXT NOT NULL,
        name TEXT,
        trashed BOOLEAN,
        created INTEGER,
        modified INTEGER,
        PRIMARY KEY (id)
    );
    ''',
    'CREATE INDEX ix_nodes_names ON nodes(name);',
    'CREATE INDEX ix_nodes_trashed ON nodes(trashed);',
    'CREATE INDEX ix_nodes_created ON nodes(created);',
    'CREATE INDEX ix_nodes_modified ON nodes(modified);',

    '''
    CREATE TABLE files (
        id TEXT NOT NULL,
        mime_type TEXT,
        md5 TEXT,
        size INTEGER,
        PRIMARY KEY (id),
        FOREIGN KEY (id) REFERENCES nodes (id)
    );
    ''',
    'CREATE INDEX ix_files_mime_type ON files(mime_type);',

    '''
    CREATE TABLE parentage (
        parent TEXT NOT NULL,
        child TEXT NOT NULL,
        PRIMARY KEY (parent, child),
        FOREIGN KEY (parent) REFERENCES nodes (id),
        FOREIGN KEY (child) REFERENCES nodes (id)
    );
    ''',
    'CREATE INDEX ix_parentage_parent ON parentage(parent);',
    'CREATE INDEX ix_parentage_child ON parentage(child);',

    '''
    CREATE TABLE images (
        id TEXT NOT NULL,
        width INTEGER NOT NULL,
        height INTEGER NOT NULL,
        PRIMARY KEY (id),
        FOREIGN KEY (id) REFERENCES nodes (id)
    );
    ''',

    '''
    CREATE TABLE videos (
        id TEXT NOT NULL,
        width INTEGER NOT NULL,
        height INTEGER NOT NULL,
        ms_duration INTEGER NOT NULL,
        PRIMARY KEY (id),
        FOREIGN KEY (id) REFERENCES nodes (id)
    );
    ''',

    'PRAGMA user_version = 3;',
]


CURRENT_SCHEMA_VERSION = 3


class CacheError(u.GoogleDriveError):

    def __init__(self, message: Text) -> None:
        super(CacheError, self).__init__()

        self._message = message

    def __str__(self) -> Text:
        return self._message


class Cache(object):

    def __init__(self, dsn: Text) -> None:
        self._dsn = dsn
        self._pool = None
        self._raii = None

    async def __aenter__(self) -> 'Cache':
        with cl.ExitStack() as stack:
            self._pool = stack.enter_context(cf.ProcessPoolExecutor())
            await self._bg(initialize)
            self._raii = stack.pop_all()
        return self

    async def __aexit__(self, type_, value, traceback) -> bool:
        self._raii.close()
        self._pool = None
        self._raii = None

    async def get_root_id(self) -> Text:
        return await self.get_metadata('root_id')

    async def get_root_node(self) -> 'Node':
        root_id = await self.get_root_id()
        return await self.get_node_by_id(root_id)

    async def get_metadata(self, key: Text) -> Text:
        return await self._bg(get_metadata, key)

    async def set_metadata(self, key: Text, value: Text) -> None:
        return await self._bg(set_metadata, key, value)

    async def get_node_by_id(self, node_id: Text) -> Union['Node', None]:
        return await self._bg(get_node_by_id, node_id)

    async def get_node_by_path(self, path: Text) -> Union['Node', None]:
        path = pl.Path(path)
        parts = list(path.parts)
        if parts[0] != '/':
            raise ValueError('invalid path')

        root_id = await self.get_root_id()
        parts.pop(0)
        return await self._bg(get_node_by_path, root_id, parts)

    async def get_path_by_id(self, node_id: Text) -> Union[Text, None]:
        return await self._bg(get_path_by_id, node_id)

    async def get_node_by_name_from_parent_id(self,
        name: Text,
        parent_id: Text
    ) -> Union['Node', None]:
        return await self._bg(get_node_by_name_from_parent_id, name, parent_id)

    async def get_children_by_id(self, node_id: Text) -> List['Node']:
        return await self._bg(get_children_by_id, node_id)

    async def apply_changes(self,
        changes: Dict[Text, Any],
        check_point: Text,
    ) -> None:
        return await self._bg(apply_changes, changes, check_point)

    async def insert_node(self, node: 'Node') -> None:
        return await self._bg(insert_node, node)

    async def find_nodes_by_regex(self, pattern: Text) -> List['Node']:
        return await self._bg(find_nodes_by_regex, pattern)

    async def find_duplicate_nodes(self) -> List['Node']:
        return await self._bg(find_duplicate_nodes)

    async def find_orphan_nodes(self) -> List['Node']:
        return await self._bg(find_orphan_nodes)

    async def find_multiple_parents_nodes(self) -> List['Node']:
        return await self._bg(find_multiple_parents_nodes)

    async def _bg(self, fn, *args):
        loop = asyncio.get_running_loop()
        return await loop.run_in_executor(self._pool, fn, self._dsn, *args)


class Node(object):

    def __init__(self,
        id_: Text,
        name: Text,
        trashed: bool,
        created: arrow.Arrow,
        modified: arrow.Arrow,
        parent_list: List[Text],
        is_folder: bool,
        mime_type: Text,
        md5: Text,
        size: int,
        image: Dict[Text, Any],
        video: Dict[Text, Any],
    ) -> None:
        self._id = id_
        self._name = name
        self._trashed = trashed
        self._created = created
        self._modified = modified
        self._parent_list = parent_list
        self._is_folder = is_folder
        self._mime_type = mime_type
        self._md5 = md5
        self._size = size
        self._image = image
        self._video = video

    def __repr__(self):
        return f"Node(id='{self.id_}')"

    @property
    def is_root(self) -> bool:
        return self._name is None

    @property
    def id_(self) -> Text:
        return self._id

    @property
    def name(self) -> Text:
        return self._name

    @property
    def trashed(self) -> bool:
        return self._trashed

    @trashed.setter
    def trashed(self, trashed: bool) -> None:
        self._trashed = trashed

    @property
    def created(self) -> arrow.Arrow:
        return self._created

    @property
    def modified(self) -> arrow.Arrow:
        return self._modified

    @property
    def parent_list(self) -> List[Text]:
        return self._parent_list

    @property
    def parent_id(self) -> Text:
        return None if not self._parent_list else self._parent_list[0]

    @property
    def is_file(self) -> bool:
        return not self._is_folder

    @property
    def is_folder(self) -> bool:
        return self._is_folder

    @property
    def mime_type(self) -> Optional[Text]:
        return self._mime_type

    @property
    def md5(self) -> Optional[Text]:
        return self._md5

    @property
    def size(self) -> Optional[int]:
        return self._size

    @property
    def is_image(self) -> bool:
        return self._image is not None

    @property
    def image_width(self) -> Optional[int]:
        return self._image['width'] if self.is_image else None

    @property
    def image_height(self) -> Optional[int]:
        return self._image['height'] if self.is_image else None

    @property
    def is_video(self) -> bool:
        return self._video is not None

    @property
    def video_width(self) -> Optional[int]:
        return self._video['width'] if self.is_video else None

    @property
    def video_height(self) -> Optional[int]:
        return self._video['height'] if self.is_video else None

    @property
    def video_ms_duration(self) -> Optional[int]:
        return self._video['ms_duration'] if self.is_video else None


def dict_from_api(data: Dict[Text, Any]) -> Dict[Text, Any]:
    id_ = data['id']
    is_folder = data['mimeType'] == u.FOLDER_MIME_TYPE
    image = None
    if 'imageMediaMetadata' in data:
        image = {
            'width': data['imageMediaMetadata']['width'],
            'height': data['imageMediaMetadata']['height'],
        }
    video = None
    if 'videoMediaMetadata' in data:
        video = {
            'width': data['videoMediaMetadata']['width'],
            'height': data['videoMediaMetadata']['height'],
            'ms_duration': data['videoMediaMetadata']['durationMillis'],
        }
    return {
        'id': id_,
        'name': data['name'],
        'trashed': data['trashed'],
        'created': data['createdTime'],
        'modified': data['modifiedTime'],
        'parent_list': data.get('parents', None),
        'is_folder': is_folder,
        'mime_type': None if is_folder else data['mimeType'],
        'md5': data.get('md5Checksum', None),
        'size': data.get('size', None),
        'image': image,
        'video': video,
    }


def node_from_api(data: Dict[Text, Any]) -> Node:
    data = dict_from_api(data)
    return node_from_database(data)


def node_from_database(data: Dict[Text, Any]) -> Node:
    return Node(
        id_=data['id'],
        name=data['name'],
        trashed=bool(data['trashed']),
        created=arrow.get(data['created']),
        modified=arrow.get(data['modified']),
        parent_list=data.get('parent_list', None),
        is_folder=data['is_folder'],
        mime_type=data['mime_type'],
        md5=data['md5'],
        size=data['size'],
        image=data['image'],
        video=data['video'],
    )


def dict_from_node(node):
    return {
        'id': node.id_,
        'parent_id': node.parent_id,
        'name': node.name,
        'trashed': node.trashed,
        'is_folder': node.is_folder,
        'created': node.created.isoformat(),
        'modified': node.modified.isoformat(),
        'mime_type': node.mime_type,
        'md5': node.md5,
        'size': node.size,
        'is_image': node.is_image,
        'image_width': node.image_width,
        'image_height': node.image_height,
        'is_video': node.is_video,
        'video_width': node.video_width,
        'video_height': node.video_height,
        'video_ms_duration': node.video_ms_duration,
    }


class Database(object):

    def __init__(self, dsn: Text) -> None:
        self._dsn = dsn

    def __enter__(self) -> sqlite3.Connection:
        self._db = sqlite3.connect(self._dsn)
        self._db.row_factory = sqlite3.Row
        return self._db

    def __exit__(self, type_, value, traceback) -> bool:
        self._db.close()


class ReadOnly(object):

    def __init__(self, db: sqlite3.Connection) -> None:
        self._db = db

    def __enter__(self) -> sqlite3.Cursor:
        self._cursor = self._db.cursor()
        return self._cursor

    def __exit__(self, type_, value, traceback) -> bool:
        self._cursor.close()


class ReadWrite(object):

    def __init__(self, db: sqlite3.Connection) -> None:
        self._db = db

    def __enter__(self) -> sqlite3.Cursor:
        self._cursor = self._db.cursor()
        return self._cursor

    def __exit__(self, type_, value, traceback) -> bool:
        if type_ is None:
            self._db.commit()
        else:
            self._db.rollback()
        self._cursor.close()


def initialize(dsn: Text):
    with Database(dsn) as db:
        try:
            # initialize table
            with ReadWrite(db) as query:
                for sql in SQL_CREATE_TABLES:
                    query.execute(sql)
        except sqlite3.OperationalError as e:
            pass

        # check the schema version
        with ReadOnly(db) as query:
            query.execute('PRAGMA user_version;')
            rv = query.fetchone()
        version = int(rv[0])

        if version < 3:
            raise CacheError((
                'impossible to migrate from old schema prior to'
                ' version 3, please rebuild the cache'
            ))


def get_metadata(dsn: Text, key: Text) -> Text:
    with Database(dsn) as db, \
         ReadOnly(db) as query:
        query.execute('SELECT value FROM metadata WHERE key = ?;', (key,))
        rv = query.fetchone()
    if not rv:
        raise KeyError
    return rv['value']


def set_metadata(dsn: Text, key: Text, value: Text) -> None:
    with Database(dsn) as db, \
         ReadWrite(db) as query:
        inner_set_metadata(query, key, value)


def get_node_by_id(dsn: Text, node_id: Text) -> Union['Node', None]:
    with Database(dsn) as db, \
         ReadOnly(db) as query:
        return inner_get_node_by_id(query, node_id)


def get_node_by_path(
    dsn: Text,
    root_id: Text,
    path_parts: List[Text],
) -> Union['Node', None]:
    node_id = root_id
    parts = path_parts
    with Database(dsn) as db, \
         ReadOnly(db) as query:
        for part in parts:
            query.execute('''
                SELECT nodes.id AS id
                FROM parentage
                    INNER JOIN nodes ON parentage.child=nodes.id
                WHERE parentage.parent=? AND nodes.name=?
            ;''', (node_id, part))
            rv = query.fetchone()
            if not rv:
                return None
            node_id = rv['id']

        node = inner_get_node_by_id(query, node_id)
    return node


def get_path_by_id(dsn: Text, node_id: Text) -> Union[Text, None]:
    parts = []
    with Database(dsn) as db, \
         ReadOnly(db) as query:
        while True:
            query.execute('''
                SELECT name
                FROM nodes
                WHERE id=?
            ;''', (node_id,))
            rv = query.fetchone()
            if not rv:
                return None

            name = rv['name']
            if not name:
                parts.insert(0, '/')
                break
            parts.insert(0, name)

            query.execute('''
                SELECT parent
                FROM parentage
                WHERE child=?
            ;''', (node_id,))
            rv = query.fetchone()
            if not rv:
                # orphan node
                break
            node_id = rv['parent']

    path = pl.Path(*parts)
    return str(path)


def get_node_by_name_from_parent_id(
    dsn: Text,
    name: Text,
    parent_id: Text
) -> Union['Node', None]:
    with Database(dsn) as db, \
         ReadOnly(db) as query:
        query.execute('''
            SELECT nodes.id AS id
            FROM nodes
                INNER JOIN parentage ON parentage.child=nodes.id
            WHERE parentage.parent=? AND nodes.name=?
        ;''', (parent_id, name))
        rv = query.fetchone()

        if not rv:
            return None

        node = inner_get_node_by_id(query, rv['id'])
    return node


def get_children_by_id(dsn: Text, node_id: Text) -> List['Node']:
    with Database(dsn) as db, \
         ReadOnly(db) as query:
        query.execute('''
            SELECT child
            FROM parentage
            WHERE parent=?
        ;''', (node_id,))
        rv = query.fetchall()

        children = [inner_get_node_by_id(query, _['child']) for _ in rv]
    return children


def apply_changes(
    dsn: Text,
    changes: Dict[Text, Any],
    check_point: Text,
) -> None:
    with Database(dsn) as db, \
         ReadWrite(db) as query:
        for change in changes:
            is_removed = change['removed']
            if is_removed:
                inner_delete_node_by_id(query, change['fileId'])
                continue

            file_ = change['file']
            is_shared = file_['shared']
            is_owned_by_me = file_['ownedByMe']
            if is_shared or not is_owned_by_me:
                continue

            node = node_from_api(file_)
            inner_insert_node(query, node)

        inner_set_metadata(query, 'check_point', check_point)


def insert_node(dsn: Text, node: 'Node') -> None:
    with Database(dsn) as db, \
         ReadWrite(db) as query:
        inner_insert_node(query, node)

        if not node.name:
            inner_set_metadata(query, 'root_id', node.id_)


def find_nodes_by_regex(dsn: Text, pattern: Text) -> List['Node']:
    with Database(dsn) as db, \
         ReadOnly(db) as query:
        fn = ft.partial(sqlite3_regexp, re.compile(pattern, re.I))
        db.create_function('REGEXP', 2, fn)
        query.execute('SELECT id FROM nodes WHERE name REGEXP ?;', (pattern,))
        rv = query.fetchall()
        rv = [inner_get_node_by_id(query, _['id']) for _ in rv]
    return rv


def find_duplicate_nodes(dsn: Text) -> List['Node']:
    with Database(dsn) as db, \
         ReadOnly(db) as query:
        query.execute('''
            SELECT nodes.name AS name, parentage.parent as parent_id
            FROM nodes
                INNER JOIN parentage ON parentage.child=nodes.id
            GROUP BY parentage.parent, nodes.name
            HAVING COUNT(*) > 1
        ;''')
        name_parent_pair = query.fetchall()

        node_id_list = []
        for pair in name_parent_pair:
            query.execute('''
                SELECT nodes.id AS id
                FROM nodes
                    INNER JOIN parentage ON parentage.child=nodes.id
                WHERE parentage.parent=? AND nodes.name=?
            ;''', (pair['parent_id'], pair['name']))
            rv = query.fetchall()
            rv = (_['id'] for _ in rv)
            node_id_list.extend(rv)

        rv = [inner_get_node_by_id(query, _) for _ in node_id_list]
    return rv


def find_orphan_nodes(dsn: Text) -> List['Node']:
    with Database(dsn) as db, \
         ReadOnly(db) as query:
        query.execute('''
            SELECT nodes.id AS id
            FROM parentage
                LEFT OUTER JOIN nodes ON parentage.child=nodes.id
            WHERE parentage.parent IS NULL
        ;''')
        rv = query.fetchall()
        rv = [inner_get_node_by_id(query, _['id']) for _ in rv]
    return rv


def find_multiple_parents_nodes(dsn: Text) -> List['Node']:
    with Database(dsn) as db, \
         ReadOnly(db) as query:
        query.execute('''
            SELECT child, COUNT(child) AS parent_count
            FROM parentage
            GROUP BY child
            HAVING parent_count > 1
        ;''')
        rv = query.fetchall()
        rv = [inner_get_node_by_id(query, _['child']) for _ in rv]
    return rv


def inner_set_metadata(query: Text, key: Text, value: Text) -> None:
    query.execute('''
        INSERT OR REPLACE INTO metadata
        VALUES (?, ?)
    ;''', (key, value))


def inner_get_node_by_id(
    query: sqlite3.Cursor,
    node_id: Text,
) -> Union['Node', None]:
    query.execute('''
        SELECT name, trashed, created, modified
        FROM nodes
        WHERE id=?
    ;''', (node_id,))
    rv = query.fetchone()
    if not rv:
        return None
    node = dict(rv)
    node['id'] = node_id

    query.execute('''
        SELECT mime_type, md5, size
        FROM files
        WHERE id=?
    ;''', (node_id,))
    rv = query.fetchone()
    is_folder = rv is None
    node['is_folder'] = is_folder
    node['mime_type'] = None if is_folder else rv['mime_type']
    node['md5'] = None if is_folder else rv['md5']
    node['size'] = None if is_folder else rv['size']

    query.execute('''
        SELECT parent
        FROM parentage
        WHERE child=?
    ;''', (node_id,))
    rv = query.fetchall()
    node['parent_list'] = None if not rv else [_['parent'] for _ in rv]

    query.execute('''
        SELECT width, height
        FROM images
        WHERE id=?
    ;''', (node_id,))
    rv = query.fetchone()
    node['image'] = {
        'width': rv['width'],
        'height': rv['height'],
    } if rv else None

    query.execute('''
        SELECT width, height, ms_duration
        FROM videos
        WHERE id=?
    ;''', (node_id,))
    rv = query.fetchone()
    node['video'] = {
        'width': rv['width'],
        'height': rv['height'],
        'ms_duration': rv['ms_duration'],
    } if rv else None

    node = node_from_database(node)
    return node


def inner_insert_node(query: sqlite3.Cursor, node: Node) -> None:
    # add this node
    query.execute('''
        INSERT OR REPLACE INTO nodes
        (id, name, trashed, created, modified)
        VALUES
        (?, ?, ?, ?, ?)
    ;''', (node.id_, node.name, node.trashed,
           node.created.timestamp, node.modified.timestamp))

    # add file information
    if not node.is_folder:
        query.execute('''
            INSERT OR REPLACE INTO files
            (id, mime_type, md5, size)
            VALUES
            (?, ?, ?, ?)
        ;''', (node.id_, node.mime_type, node.md5, node.size))

    # remove old parentage
    query.execute('''
        DELETE FROM parentage
        WHERE child=?
    ;''', (node.id_,))
    # add parentage if there is any
    if node.parent_list:
        for parent in node.parent_list:
            query.execute('''
                INSERT INTO parentage
                (parent, child)
                VALUES
                (?, ?)
            ;''', (parent, node.id_))

    # add image information
    if node.is_image:
        query.execute('''
            INSERT OR REPLACE INTO images
            (id, width, height)
            VALUES
            (?, ?, ?)
        ;''', (node.id_, node.image_width, node.image_height))

    # add video information
    if node.is_video:
        query.execute('''
            INSERT OR REPLACE INTO videos
            (id, width, height, ms_duration)
            VALUES
            (?, ?, ?, ?)
        ;''', (node.id_, node.video_width, node.video_height,
               node.video_ms_duration))


def inner_delete_node_by_id(query: sqlite3.Cursor, node_id: Text) -> None:
    # remove from videos
    query.execute('''
        DELETE FROM videos
        WHERE id=?
    ;''', (node_id,))

    # remove from images
    query.execute('''
        DELETE FROM images
        WHERE id=?
    ;''', (node_id,))

    # disconnect parents
    query.execute('''
        DELETE FROM parentage
        WHERE child=? OR parent=?
    ;''', (node_id, node_id))

    # remove from files
    query.execute('''
        DELETE FROM files
        WHERE id=?
    ;''', (node_id,))

    # remove from nodes
    query.execute('''
        DELETE FROM nodes
        WHERE id=?
    ;''', (node_id,))


def sqlite3_regexp(
    pattern: 'Pattern',
    _: Text,
    cell: Union[Text, None],
) -> bool:
    if cell is None:
        # root node
        return False
    return pattern.search(cell) is not None



import os.path as op

from setuptools import setup


with open(op.join(op.dirname(__file__), './README.md')) as fin:
    long_description = fin.read()


setup(
        name='wcpan.drive.google',
        version='4.0.2',
        author='Wei-Cheng Pan',
        author_email='legnaleurc@gmail.com',
        description='Asynchronous Google Drive API.',
        long_description=long_description,
        long_description_content_type='text/markdown',
        url='https://github.com/legnaleurc/wcpan.drive.google',
        packages=[
            'wcpan.drive.google',
        ],
        python_requires='>= 3.7',
        install_requires=[
            'PyYAML ~= 3.12',
            'aiohttp ~= 3.4.4',
            'arrow ~= 0.12.1',
            'wcpan.logger ~= 1.2.3',
            'wcpan.worker ~= 4.1.2',
        ],
        classifiers=[
            'Programming Language :: Python :: 3 :: Only',
            'Programming Language :: Python :: 3.7',
        ])
